/*-
 * Copyright (C)2012..2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2012..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: octets.c,v 1.4 2016/02/19 08:37:45 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>

#define CRC32_INIT  0xFFFFFFFF
#define CRC32_POLY  0xEDB88320
#define CRC24_INIT  0x00B704CE
#define CRC24_POLY  0x01864CFB

const u_int8_t
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_cnt1[256] =                                                                               /****
 *****************************************************************************************************
 *****************************************************************************************************/
{ 0, 1, 1, 2,  1, 2, 2, 3,  1, 2, 2, 3,  2, 3, 3, 4
, 1, 2, 2, 3,  2, 3, 3, 4,  2, 3, 3, 4,  3, 4, 4, 5
, 1, 2, 2, 3,  2, 3, 3, 4,  2, 3, 3, 4,  3, 4, 4, 5
, 2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6

, 1, 2, 2, 3,  2, 3, 3, 4,  2, 3, 3, 4,  3, 4, 4, 5
, 2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6
, 2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6
, 3, 4, 4, 5,  4, 5, 5, 6,  4, 5, 5, 6,  5, 6, 6, 7

, 1, 2, 2, 3,  2, 3, 3, 4,  2, 3, 3, 4,  3, 4, 4, 5
, 2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6
, 2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6
, 3, 4, 4, 5,  4, 5, 5, 6,  4, 5, 5, 6,  5, 6, 6, 7

, 2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6
, 3, 4, 4, 5,  4, 5, 5, 6,  4, 5, 5, 6,  5, 6, 6, 7
, 3, 4, 4, 5,  4, 5, 5, 6,  4, 5, 5, 6,  5, 6, 6, 7
, 4, 5, 5, 6,  5, 6, 6, 7,  5, 6, 6, 7,  6, 7, 7, 8
};

u_int32_t                     /* IEEE 802.3 */
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_crc32(const u_char *octets, ssize_t len) {                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    u_int32_t carry, r;
    ssize_t i;
    int j;

    r = CRC32_INIT;
    if  (len < 0) {
        for (i = -len; i > 0;) {
            r ^= octets[--i];
            for (j = 0; j < 8; ++j) {
                carry = r & 1;
                r >>= 1;
                if  (carry) r ^= CRC32_POLY;
        }   }
    } else {
        for (i = 0; i < len; ++i) {
            r ^= octets[i];
            for (j = 0; j < 8; ++j) {
                carry = r & 1;
                r >>= 1;
                if  (carry) r ^= CRC32_POLY;
    }   }   }
    r ^= CRC32_INIT;
    return(r);
}

u_int32_t                       /* RFC-2440 */
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_crc24(const u_char *octets, ssize_t len) {                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    u_int32_t r;
    ssize_t i;
    int j;

    r = CRC24_INIT;
    if  (len < 0) {
        for (i = -len; i > 0;) {
            r ^= octets[--i] << 16;
            for (j = 0; j < 8; ++j)  {
                r <<= 1;
                if  (r & 0x1000000) r ^= CRC24_POLY;
        }   }
    } else {
        for (i = 0; i < len; ++i) {
            r ^= octets[i] << 16;
            for (j = 0; j < 8; ++j)  {
                r <<= 1;
                if  (r & 0x1000000) r ^= CRC24_POLY;
    }   }   }
    r &= 0x00FFFFFF;
    return(r);
}

u_int32_t
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_csum1(const u_char *octets, ssize_t len) {                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    u_int32_t r;
    ssize_t i;

    r = 0;
    if  (0 > len) len = -len;
    for (i = 0; i < len; ++i) {
        r += octets[i];
    }
    --r;
    r &= 0x000000FF;
    return(r);
}

u_int64_t
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_load(const u_char *octets, ssize_t len) {                                                 /****
 *****************************************************************************************************
 *****************************************************************************************************/
    u_int64_t r;
    ssize_t i;

    r = 0;
    if  (len < 0) {
        for (i = 0; i < -len; ++i) {
            r <<= 8;
            r |= octets[i];
        }
    } else {
        for (i = len; i > 0;) {
            r <<= 8;
            r |= octets[--i];
    }   }
    return(r);
}

void
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_store(u_char *octets, ssize_t len, u_int64_t c) {                                         /****
 *****************************************************************************************************
 *****************************************************************************************************/
    ssize_t i;

    if  (len < 0) {
        for (i = 0; i > len; --i) {
            octets[i - len - 1] = c & 0xFF;
            c >>= 8;
        }
    } else {
        for (i = 0; i < len; ++i) {
            octets[i] = c & 0xFF;
            c >>= 8;
}   }   }

u_int64_t
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_loadM(const u_char *octets, const char *mask) {                                           /****
 *****************************************************************************************************
 *****************************************************************************************************/
    u_int64_t r = 0;
    ssize_t i;

    for (i = 0; !!mask[i]; ++i) {
        if  ((mask[i] < '0') || (mask[i] > '7')) {
            ifBLIN_QX0("Illegal input %c in %s[%d]", mask[i], mask, i);
            errno = EINVAL;
            break;
        }
        r |= octets[i] << ((mask[i] - '0') << 3);
    }
    return(r);
}

void
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_storeM(u_char *octets, const char *mask, u_int64_t c) {                                   /****
 *****************************************************************************************************
 *****************************************************************************************************/
    ssize_t i;

    for (i = 0; !!mask[i]; ++i) {
        if  ((mask[i] < '0') || (mask[i] > '7')) {
            ifBLIN_QX0("Illegal input %c in %s[%d]", mask[i], mask, i);
            errno = EINVAL;
            break;
        }
        octets[i] = c >> ((mask[i] - '0') << 3);
}   }

u_int64_t
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_swap(u_int64_t c, u_int8_t m) {                                                           /****
 *****************************************************************************************************
 *****************************************************************************************************/
    if  (m & 0x01) c = ((c >>  1) & 0x5555555555555555ULL) | ((c <<  1) & 0xAAAAAAAAAAAAAAAAULL);
    if  (m & 0x02) c = ((c >>  2) & 0x3333333333333333ULL) | ((c <<  2) & 0xCCCCCCCCCCCCCCCCULL);
    if  (m & 0x04) c = ((c >>  4) & 0x0F0F0F0F0F0F0F0FULL) | ((c <<  4) & 0xF0F0F0F0F0F0F0F0ULL);
    if  (m & 0x08) c = ((c >>  8) & 0x00FF00FF00FF00FFULL) | ((c <<  8) & 0xFF00FF00FF00FF00ULL);
    if  (m & 0x10) c = ((c >> 16) & 0x0000FFFF0000FFFFULL) | ((c << 16) & 0xFFFF0000FFFF0000ULL);
    if  (m & 0x20) c = ((c >> 32) & 0x00000000FFFFFFFFULL) | ((c << 32) & 0xFFFFFFFF00000000ULL);
    return(c);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_count(u_int64_t c) {                                                                      /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int i;

    for (i = 0; !!c; c >>= 8) {
        i += recobe_cnt1[c & 0xFF];
    }
    return(i);
}
